import tensorflow as tf
from tensorflow.keras.models import Model
from tensorflow.keras.layers import Conv1D, MaxPooling1D, BatchNormalization, Dropout, Dense, Flatten, Input, Add, ReLU
from tensorflow.keras.optimizers import Adam


def create_model_wide_deep(signal_len, features_len, LR, n_outputs):
    print('Learning Rate is: ', LR)

    signal_input = Input(shape=(signal_len, 1), name='signal_input')

    # DEEP (1DCNN) Network
    CNN_0 = Conv1D(filters=64, kernel_size=7, activation='relu')(signal_input)
    CNN_1 = Conv1D(filters=64, kernel_size=7, activation='relu')(CNN_0)
    CNN_2 = Conv1D(filters=64, kernel_size=7, activation='relu')(CNN_1)
    drop_1 = Dropout(0.3)(CNN_2)
    pool_1 = MaxPooling1D(pool_size=2)(drop_1)
    bn_1 = BatchNormalization()(pool_1)
    CNN_3 = Conv1D(filters=128, kernel_size=7, activation='relu')(bn_1)
    CNN_4 = Conv1D(filters=128, kernel_size=7, activation='relu')(CNN_3)
    drop_2 = Dropout(0.3)(CNN_4)
    pool_2 = MaxPooling1D(pool_size=2)(drop_2)
    bn_2 = BatchNormalization()(pool_2)
    CNN_5 = Conv1D(filters=256, kernel_size=7, activation='relu')(bn_2)
    CNN_6 = Conv1D(filters=256, kernel_size=7, activation='relu')(CNN_5)
    drop_3 = Dropout(0.3)(CNN_6)
    pool_3 = MaxPooling1D(pool_size=2)(drop_3)
    bn_3 = BatchNormalization()(pool_3)
    CNN_7 = Conv1D(filters=256, kernel_size=7, activation='relu')(bn_3)
    CNN_8 = Conv1D(filters=256, kernel_size=7, activation='relu')(CNN_7)
    drop_4 = Dropout(0.3)(CNN_8)
    pool_4 = MaxPooling1D(pool_size=2)(drop_4)
    bn_4 = BatchNormalization()(pool_4)
    flat = Flatten()(bn_4)
    dense1 = Dense(500, activation='relu')(flat)
    deep_output = Dense(100, activation='relu')(dense1)

    deep_model = Model(inputs=signal_input, outputs=deep_output)

    # SHALLOW (MLP) Network
    shallow_input = Input(features_len, name='feature_input')
    shallow_output = Dense(6, activation='relu', kernel_regularizer=tf.keras.regularizers.l2(l=0.1))(shallow_input)
    shallow_model = Model(inputs=shallow_input, outputs=shallow_output)

    # Combine the shallow and deep network outputs
    concat_layer = tf.keras.layers.concatenate([deep_model.outputs[0], shallow_model.outputs[0]])
    dense2 = Dense(40, activation='relu')(concat_layer)
    output = Dense(n_outputs, activation='sigmoid')(dense2)

    # Create and compile the model
    model = Model(inputs=[signal_input, shallow_input], outputs=output)
    opt = Adam(learning_rate=LR, amsgrad=True)
    model.compile(loss='binary_crossentropy', optimizer=opt, metrics=['accuracy'])
    return model


def create_model_wide_deep_small(signal_len, features_len, LR, n_outputs):
    print('Learning Rate is: ', LR)

    signal_input = Input(shape=(signal_len, 1), name='signal_input')

    # DEEP (1DCNN) Network
    CNN_0 = Conv1D(filters=16, kernel_size=13, activation='relu')(signal_input)
    drop_1 = Dropout(0.2)(CNN_0)
    pool_1 = MaxPooling1D(pool_size=2)(drop_1)
    bn_1 = BatchNormalization()(pool_1)
    CNN_2 = Conv1D(filters=32, kernel_size=11, activation='relu')(bn_1)
    drop_2 = Dropout(0.2)(CNN_2)
    pool_2 = MaxPooling1D(pool_size=2)(drop_2)
    bn_2 = BatchNormalization()(pool_2)
    CNN_3 = Conv1D(filters=64, kernel_size=9, activation='relu')(bn_2)
    drop_3 = Dropout(0.2)(CNN_3)
    pool_3 = MaxPooling1D(pool_size=2)(drop_3)
    bn_3 = BatchNormalization()(pool_3)
    flat = Flatten()(bn_3)
    dense1 = Dense(300, activation='relu')(flat)
    deep_output = Dense(50, activation='relu')(dense1)

    deep_model = Model(inputs=signal_input, outputs=deep_output)

    # SHALLOW (MLP) Network
    shallow_input = Input(features_len, name='feature_input')
    shallow_output = Dense(10, activation='relu', kernel_regularizer=tf.keras.regularizers.l2(l=0.1))(shallow_input)
    shallow_model = Model(inputs=shallow_input, outputs=shallow_output)

    # Combine the shallow and deep network outputs
    concat_layer = tf.keras.layers.concatenate([deep_model.outputs[0], shallow_model.outputs[0]])
    dense2 = Dense(20, activation='relu')(concat_layer)
    output = Dense(n_outputs, activation='sigmoid')(dense2)

    # Create and compile the model
    model = Model(inputs=[signal_input, shallow_input], outputs=output)
    opt = Adam(learning_rate=LR, amsgrad=True)
    model.compile(loss='binary_crossentropy', optimizer=opt, metrics=['accuracy'])
    return model


def create_model_deep(signal_len, LR, n_outputs):
    print('Learning Rate is: ', LR)

    signal_input = Input(shape=(signal_len, 1))

    # DEEP (1DCNN) Network
    CNN_0 = Conv1D(filters=16, kernel_size=13, activation='relu')(signal_input)
    CNN_1 = Conv1D(filters=16, kernel_size=13, activation='relu')(CNN_0)
    drop_1 = Dropout(0.3)(CNN_1)
    pool_1 = MaxPooling1D(pool_size=2)(drop_1)
    bn_1 = BatchNormalization()(pool_1)
    CNN_2 = Conv1D(filters=32, kernel_size=9, activation='relu')(bn_1)
    CNN_3 = Conv1D(filters=32, kernel_size=9, activation='relu')(CNN_2)
    drop_2 = Dropout(0.3)(CNN_3)
    pool_2 = MaxPooling1D(pool_size=2)(drop_2)
    bn_2 = BatchNormalization()(pool_2)
    CNN_4 = Conv1D(filters=64, kernel_size=7, activation='relu')(bn_2)
    CNN_5 = Conv1D(filters=64, kernel_size=7, activation='relu')(CNN_4)
    drop_3 = Dropout(0.3)(CNN_5)
    pool_3 = MaxPooling1D(pool_size=2)(drop_3)
    bn_3 = BatchNormalization()(pool_3)
    flat = Flatten()(bn_3)
    dense1 = Dense(300, activation='relu')(flat)
    dense2 = Dense(50, activation='relu')(dense1)
    output = Dense(n_outputs, activation='sigmoid')(dense2)

    # Create and compile the model
    model = Model(inputs=signal_input, outputs=output)
    opt = Adam(learning_rate=LR, amsgrad=True)
    model.compile(loss='binary_crossentropy', optimizer=opt, metrics=['accuracy'])
    return model


def create_model_deep2(signal_len, LR, n_outputs):
    print('Learning Rate is: ', LR)

    signal_input = Input(shape=(signal_len, 1))

    # DEEP (1DCNN) Network
    CNN_0 = Conv1D(filters=16, kernel_size=13, activation='relu')(signal_input)
    CNN_1 = Conv1D(filters=16, kernel_size=13, activation='relu')(CNN_0)
    drop_1 = Dropout(0.3)(CNN_1)
    pool_1 = MaxPooling1D(pool_size=2)(drop_1)
    bn_1 = BatchNormalization()(pool_1)

    CNN_2 = Conv1D(filters=32, kernel_size=11, activation='relu')(bn_1)
    CNN_3 = Conv1D(filters=32, kernel_size=11, activation='relu')(CNN_2)
    drop_2 = Dropout(0.3)(CNN_3)
    pool_2 = MaxPooling1D(pool_size=2)(drop_2)
    bn_2 = BatchNormalization()(pool_2)

    CNN_4 = Conv1D(filters=64, kernel_size=9, activation='relu')(bn_2)
    CNN_5 = Conv1D(filters=64, kernel_size=9, activation='relu')(CNN_4)
    drop_3 = Dropout(0.3)(CNN_5)
    pool_3 = MaxPooling1D(pool_size=2)(drop_3)
    bn_3 = BatchNormalization()(pool_3)

    CNN_6 = Conv1D(filters=64, kernel_size=7, activation='relu')(bn_3)
    CNN_7 = Conv1D(filters=64, kernel_size=7, activation='relu')(CNN_6)
    drop_4 = Dropout(0.3)(CNN_7)
    pool_4 = MaxPooling1D(pool_size=2)(drop_4)
    bn_4 = BatchNormalization()(pool_4)

    flat = Flatten()(bn_4)
    dense1 = Dense(300, activation='relu')(flat)
    dense2 = Dense(50, activation='relu')(dense1)
    output = Dense(n_outputs, activation='sigmoid')(dense2)

    # Create and compile the model
    model = Model(inputs=signal_input, outputs=output)
    opt = Adam(learning_rate=LR, amsgrad=True)
    model.compile(loss='binary_crossentropy', optimizer=opt, metrics=['accuracy'])
    return model


def create_model_deep_small(signal_len, LR, n_outputs):
    print('Learning Rate is: ', LR)

    signal_input = Input(shape=(signal_len, 1))

    # DEEP (1DCNN) Network
    CNN_0 = Conv1D(filters=16, kernel_size=11, activation='relu')(signal_input)
    drop_1 = Dropout(0.1)(CNN_0)
    pool_1 = MaxPooling1D(pool_size=2)(drop_1)
    bn_1 = BatchNormalization()(pool_1)
    CNN_2 = Conv1D(filters=32, kernel_size=9, activation='relu')(bn_1)
    drop_2 = Dropout(0.1)(CNN_2)
    pool_2 = MaxPooling1D(pool_size=2)(drop_2)
    bn_2 = BatchNormalization()(pool_2)
    CNN_3 = Conv1D(filters=64, kernel_size=7, activation='relu')(bn_2)
    drop_3 = Dropout(0.1)(CNN_3)
    pool_3 = MaxPooling1D(pool_size=2)(drop_3)
    bn_3 = BatchNormalization()(pool_3)
    flat = Flatten()(bn_3)
    dense1 = Dense(300, activation='relu')(flat)
    dense2 = Dense(50, activation='relu')(dense1)
    output = Dense(n_outputs, activation='sigmoid')(dense2)

    # Create and compile the model
    model = Model(inputs=signal_input, outputs=output)
    opt = Adam(learning_rate=LR, amsgrad=True)
    model.compile(loss='binary_crossentropy', optimizer=opt, metrics=['accuracy'])
    return model
