#!/usr/bin/env python

# Edit this script to add your team's code. Some functions are *required*,
# but you can edit most parts of the required functions,
# change or remove non-required functions, and add your own functions.

# Command for training the model:
# python ./MRH_v21/train_model.py './dataset/sample_data/' 'Model_1' 2

# Command for testing the model:
# python ./MRH_v21/run_model.py 'Model_1' './dataset/sample_data/' 'OUT_folder'

# Command for evaluating the model:
# python ./MRH_v21/evaluate_model.py './dataset/sample_data/' './OUT_folder/'
################################################################################
#
# Import libraries and functions. You can change or remove them.
#
################################################################################

from helper_code import *
import numpy as np, scipy as sp, scipy.stats, os, sys
# from sklearn.impute import SimpleImputer

# Added by MRH
import time
import random
import tensorflow as tf
import pandas as pd
import pickle
from tqdm import tqdm
from sklearn.preprocessing import StandardScaler
# from sklearn.decomposition import PCA
from create_model import *

config = tf.compat.v1.ConfigProto()
config.gpu_options.per_process_gpu_memory_fraction = 0.7
config.gpu_options.allow_growth = True
tf.compat.v1.keras.backend.set_session(
    tf.compat.v1.Session(config=config))

random.seed(1234)
np.random.seed(1234)
tf.random.set_seed(1234)

LR = 5e-5  # learning rate
window_len = 4  # length of signal in second
win_shift = 2  # step size (stride) in second for overlap windows
resampled_freq = 1000
dataset_portion = 1.00
# dataset_portion = 0.01 # <<<<<<<<<<<------------*********** a portion of data is used for testing the algorithm !!!
train_portion = 0.90
num_epochs = 20
patience = 5


################################################################################
#
# Required functions. Edit these functions to add your code, but do not change the arguments.
#
################################################################################
# load signals data
def load_dataset(data_folder, window_len, win_shift, resampled_freq, verbose):
    # Loads the dataset
    # header_files, recording_files: list of files with their path
    # window_len: length of signal (e.g., 3 second)
    # win_shift: stride for overlap; if zero, then no shift
    # classes: list of classes (e.g., 3 classes as 'Present', 'Unknown', 'Absent')

    # Find data files.
    if verbose >= 1:
        print('Finding data files...')

    murmur_classes = ['Present', 'Unknown', 'Absent']
    num_murmur_classes = len(murmur_classes)
    outcome_classes = ['Abnormal', 'Normal']
    num_outcome_classes = len(outcome_classes)

    patient_files = find_patient_files(data_folder)  # return the list of .TXT files
    num_patient_files = len(patient_files)
    if verbose >= 2:
        print("number of patient files = ", num_patient_files)
    if num_patient_files == 0:
        raise Exception('No data was provided.')

    murmur_labels = list()
    outcome_labels = list()
    signals = list()
    # features = list()

    counter = 0
    for i in tqdm(range(num_patient_files)):
        # for i in range(num_recordings):
        if verbose >= 2:
            print('    {}/{}...'.format(i + 1, num_patient_files))
            # print('------------------------------------++++++++>>>>>>>>>> Name of file: {}'.format(header_files[i]))

        # Load the current patient data and recordings.
        current_patient_data = load_patient_data(patient_files[i])
        murmur_locations = get_murmur_locations(current_patient_data)
        # print('murmur_locations =', murmur_locations)
        # current_recordings = load_recordings(data_folder, current_patient_data)
        current_recordings = load_murmur_recordings(data_folder, current_patient_data)
        counter += len(current_recordings)

        # resampling signals
        # print('length of signals before resampling:')
        # for r in range(len(current_recordings)):
        #     print('{}, {}: '.format(r,len(current_recordings[r])))

        freq = get_frequency(current_patient_data)
        decimate = False
        rate = 1
        if (freq / resampled_freq) == (freq // resampled_freq):
            rate = int(freq // resampled_freq)
            if rate > 1:
                decimate = True

        for r in range(len(current_recordings)):
            if decimate:
                # print('length of signals {} before resampling:, {}'.format(r,len(current_recordings[r])))
                current_recordings[r] = current_recordings[r][::rate]
                # print('length of signals {} after resampling:, {}'.format(r,len(current_recordings[r])))
            else:
                for i in range(len(current_recordings)):
                    num_samp = len(current_recordings[i])  # length of each signals
                    # resampling using interpoation
                    samples_time = np.linspace(0, num_samp / freq, num=num_samp, endpoint=True)
                    func = interp1d(samples_time, current_recordings[i], kind='cubic', axis=1)
                    scale = resampled_freq / freq
                    new_samples_time = np.linspace(0, num_samp / freq, num=int(num_samp * scale), endpoint=True)
                    current_recordings[i] = func(new_samples_time)
                    # print('length of signals after resampling:')
        # for r in range(len(current_recordings)):
        #     print('i: ',len(current_recordings[r]))

        # Extract features.
        # current_features = get_features(current_patient_data, current_recordings)

        # Extract labels and use one-hot encoding. (NEW)
        current_murmur_labels = np.zeros(num_murmur_classes, dtype=int)
        murmur = get_murmur(current_patient_data)
        if murmur in murmur_classes:
            j = murmur_classes.index(murmur)
            current_murmur_labels[j] = 1
        # murmurs.append(current_murmur_labels)

        current_outcome_labels = np.zeros(num_outcome_classes, dtype=int)
        outcome = get_outcome(current_patient_data)
        if outcome in outcome_classes:
            j = outcome_classes.index(outcome)
            current_outcome_labels[j] = 1
        # outcomes.append(current_outcome_labels)

        # Extract signals
        for r in range(len(current_recordings)):
            j = 0
            # current_recordings[i] = nk.signal_filter(current_recordings[i], lowcut=5, highcut=100)
            winsamp = window_len * resampled_freq  # how many samples in segmented signal
            winsamp_shift = int(win_shift * resampled_freq)  # how much to shift segmentation frame
            sig = current_recordings[r]
            sig_len = len(sig)
            win = sig[j:j + winsamp]
            windowed_signal = [win]
            count = 1
            if win_shift:
                j += winsamp_shift
                while (j + winsamp < sig_len):
                    win = sig[j:j + winsamp]
                    count = count + 1
                    windowed_signal.append(win)
                    j += winsamp_shift
                windowed_signal = np.vstack(windowed_signal)
            else:  # in the case that overlap is false (win_shift = 0)
                j += winsamp
                while (j + winsamp < sig_len):
                    win = sig[j:j + winsamp]
                    count = count + 1
                    windowed_signal.append(win)
                    j += winsamp
                windowed_signal = np.vstack(windowed_signal)
            signals.append(windowed_signal)
            for i in range(count):
                # append the label to the list of labels,
                # as these labels are the same for all segmented windowed signals
                murmur_labels.append(current_murmur_labels)
                outcome_labels.append(current_outcome_labels)
                # features.append(current_features)
    signals = np.vstack(signals)
    murmur_labels = np.vstack(murmur_labels)
    outcome_labels = np.vstack(outcome_labels)
    # features = np.vstack(features)
    print('shape of signals after resampling: {}'.format(signals.shape))

    # Perform signal scalling.
    signal_scaler = StandardScaler()
    dim = signals.shape
    signals = signal_scaler.fit_transform(signals)
    # print('shape of segmented signals after StandardScaler function and reshape: {}'.format(signals.shape))
    signals = signals.reshape(dim[0], dim[1], 1)  # for feeding to the CNN model

    # feature_scaler = StandardScaler()
    # features = feature_scaler.fit_transform(features)
    # print('shape of features after StandardScaler function: {}'.format(features.shape))

    #     # turn any NaNs to 0 for PCA handling
    #     features[np.isnan(features)] = 0

    #     # create PCA transformer, and perform PCA
    #     pca = PCA(n_components = 10)
    #     features = pca.fit_transform(features)
    #     print('Type of  features after PCA function: {}'.format(type(features)))
    #     print('shape of features after PCA function: {}'.format(features.shape))
    print('---------------------------------->>> {} signals from {} patients.'.format(counter, num_patient_files))

    return signals, signal_scaler, murmur_labels, outcome_labels  # , features, feature_scaler #, pca

# Train the model.
def train_challenge_model(data_folder, model_folder, verbose):
    t = time.time()
    # Create a folder for the model if it does not already exist.
    os.makedirs(model_folder, exist_ok=True)

    # Load data (signals and features)
    # signals, signal_scaler, labels, features, feature_scaler = load_dataset(data_folder, window_len, win_shift,
    # resampled_freq, verbose)
    signals, signal_scaler, murmur_labels, outcome_labels = load_dataset(data_folder, window_len, win_shift,
                                                                         resampled_freq, verbose)

    # spliting Train and Validation datasets before shuffling to avoid overlap of segmented windows of
    # the same signal in both dataset!
    n_train = int(train_portion * len(signals))
    print('Total number of signals= ', len(signals))
    # print('Total number of training signals= ', n_train)

    signals_val = signals[n_train:]
    # print('Total number of validation signals= ', len(signals_val))
    # features_val = features[n_train:]
    murmur_labels_val = murmur_labels[n_train:]
    outcome_labels_val = outcome_labels[n_train:]
    signals = signals[:n_train]
    # print('New Total number of training signals= ', len(signals))

    # features = features[:n_train]
    murmur_labels = murmur_labels[:n_train]
    outcome_labels = outcome_labels[:n_train]

    # shuffling the train dataset and the validation dataset
    # create random indexing for shuffling the train dataset and the validation dataset
    val_idx = np.asarray(random.sample(range(len(signals_val)), len(signals_val)), dtype=np.int)
    train_idx = np.asarray(random.sample(range(n_train), n_train), dtype=np.int)
    signals_val = signals_val[val_idx]
    # print('Total number of validation signals= ', len(signals_val))
    # print('Length of validation random index= ', len(val_idx))
    # features_val = features_val[val_idx]
    murmur_labels_val = murmur_labels_val[val_idx]
    outcome_labels_val = outcome_labels_val[val_idx]
    signals = signals[train_idx]
    # features = features[train_idx]
    murmur_labels = murmur_labels[train_idx]
    outcome_labels = outcome_labels[train_idx]
    print('\nSize of training and validation datasets: {} and {}'.format(len(signals), len(signals_val)))
    # print('----------------->>>>>>>>>> length of each signal: {}'.format(signals.shape[1]))

    print('. . . Creating model!')
    # classifier = create_model_wide_deep(signals.shape[1], features.shape[1],LR)
    # classifier = create_model_wide_deep_small(signals.shape[1], features.shape[1],LR)
    # classifier.summary()
    # callback = tf.keras.callbacks.EarlyStopping(monitor='val_loss', patience=patience, restore_best_weights=True)
    # history = classifier.fit({'signal_input': signals,'feature_input': features}, labels,
    #                     validation_data = ({'signal_input': signals_val, 'feature_input': features_val}, labels_val),
    #                     batch_size = 64, epochs = num_epochs, callbacks=[callback], verbose = 1)

    murmur_classes = ['Present', 'Unknown', 'Absent']
    outcome_classes = ['Abnormal', 'Normal']

    # create deep (not wide) model
    murmur_classifier = create_model_deep2(signals.shape[1], LR,len(murmur_classes))
    print('Type of murmur_classifier: ', type(murmur_classifier))
    murmur_classifier.summary()
    callback = tf.keras.callbacks.EarlyStopping(monitor='val_loss', patience=patience, restore_best_weights=True)
    history = murmur_classifier.fit(signals, murmur_labels,
                             validation_data=(signals_val, murmur_labels_val),
                             batch_size=64, epochs=num_epochs, callbacks=[callback], verbose=1)

    outcome_classifier = create_model_deep2(signals.shape[1], LR,len(outcome_classes))
    print('Type of outcome_classifier: ', type(outcome_classifier))
    outcome_classifier.summary()
    callback = tf.keras.callbacks.EarlyStopping(monitor='val_loss', patience=patience, restore_best_weights=True)
    history = outcome_classifier.fit(signals, outcome_labels,
                             validation_data=(signals_val, outcome_labels_val),
                             batch_size=64, epochs=num_epochs, callbacks=[callback], verbose=1)

    print("Total training took " + str(round(time.time() - t, 0)) + ' seconds')

    # Save the model.
    save_challenge_model(model_folder, murmur_classes, outcome_classes, murmur_classifier, outcome_classifier,
                         signal_scaler)  # , feature_scaler

    if verbose >= 1:
        print('Done.')

# Get murmur locations from patient data.
def get_murmur_locations(data):
    locations_list = list()
    for l in data.split('\n'):
        if l.startswith('#Murmur locations:'):
            try:
                locations_list = l.split(': ')[1]
            except:
                pass
    murmur_locations = locations_list.split('+')
    return murmur_locations

# Load recordings.
def load_murmur_recordings(data_folder, data, get_frequencies=False):
    num_locations = get_num_locations(data)
    recording_information = data.split('\n')[1:num_locations+1]
    murmur_locations = get_murmur_locations(data)
    # print('murmur_locations: ',murmur_locations)
    # print('Length of murmur_locations: ',len(murmur_locations))

    recordings = list()
    frequencies = list()
    if murmur_locations != ['nan']:
        for i in range(num_locations):
            entries = recording_information[i].split(' ')
            # print('Signal location: ', entries[0])
            if entries[0] in murmur_locations:
                recording_file = entries[2]
                filename = os.path.join(data_folder, recording_file)
                recording, frequency = load_wav_file(filename)
                recordings.append(recording)
                # print('Signal with file-name {} was added!'.format(entries[2]))
                frequencies.append(frequency)
    else:
        # print('This file does not have murmur!!!!')
        for i in range(num_locations):
            entries = recording_information[i].split(' ')
            recording_file = entries[2]
            filename = os.path.join(data_folder, recording_file)
            recording, frequency = load_wav_file(filename)
            recordings.append(recording)
            # print('Signal location: ', entries[0])
            # print('Signal with file-name {} was added!'.format(entries[2]))
            frequencies.append(frequency)

    if get_frequencies:
        return recordings, frequencies
    else:
        return recordings


# Load your trained model. This function is *required*. You should edit this function to add your code,
# but do *not* change the arguments of this function.
def load_challenge_model(model_folder, verbose):
    murmur_classes = pd.read_csv(model_folder + '/' + 'murmur_classes.csv')['0'].to_list()
    outcome_classes = pd.read_csv(model_folder + '/' + 'outcome_classes.csv')['0'].to_list()
    murmur_classifier = tf.keras.models.load_model(model_folder + '/' + 'murmur_model')
    outcome_classifier = tf.keras.models.load_model(model_folder + '/' + 'outcome_model')
    signal_scaler = pickle.load(open(model_folder + '/scaler.pkl', 'rb'))['signal_scaler']
    # feature_scaler = pickle.load(open(model_folder + '/scaler.pkl', 'rb'))['feature_scaler']
    # model = {'classes': classes, 'classifier': classifier,
    #          'signal_scaler': signal_scaler, 'feature_scaler': feature_scaler}
    model = {'murmur_classes': murmur_classes, 'murmur_classifier': murmur_classifier,
             'outcome_classes': outcome_classes, 'outcome_classifier': outcome_classifier,
             'signal_scaler': signal_scaler}
    return model


# Run your trained model. This function is *required*. You should edit this function to add your code,
# but do *not* change the arguments of this function.
def run_challenge_model(model, data, current_recordings, verbose):
    murmur_classes = model['murmur_classes']
    outcome_classes = model['outcome_classes']
    murmur_classifier = model['murmur_classifier']
    outcome_classifier = model['outcome_classifier']
    signal_scaler = model['signal_scaler']
    # feature_scaler = model['feature_scaler']

    # resampling
    decimate = False
    freq = get_frequency(data)
    if (freq / resampled_freq) == (freq // resampled_freq):
        rate = int(freq // resampled_freq)
        if rate > 1:
            decimate = True

    if decimate:
        for i in range(len(current_recordings)):
            current_recordings[i] = current_recordings[i][::rate]
    else:
        for i in range(len(current_recordings)):
            num_samp = len(current_recordings[i])  # length of each signals
            # resampling using interpolation
            samples_time = np.linspace(0, num_samp / freq, num=num_samp, endpoint=True)
            func = interp1d(samples_time, current_recordings[i], kind='cubic', axis=1)
            scale = resampled_freq / freq
            new_samples_time = np.linspace(0, num_samp / freq, num=int(num_samp * scale), endpoint=True)
            current_recordings[i] = func(new_samples_time)

    # Extract features.
    # current_features = get_features(data, current_recordings)
    # print('len of features:',len(current_features))
    # features = list()

    signals = list()
    # print('length of recordings:',len(current_recordings))
    # Extract signals
    for r in range(len(current_recordings)):
        # print('loop index:',r)
        j = 0
        # current_recordings[i] = nk.signal_filter(current_recordings[i], lowcut=5, highcut=100)
        winsamp = window_len * resampled_freq  # how many samples in segmented signal
        winsamp_shift = int(win_shift * resampled_freq)  # how much to shift segmentation frame
        sig = current_recordings[r]
        sig_len = len(sig)
        win = sig[j:j + winsamp]
        windowed_signal = [win]
        count = 1
        if win_shift:
            j += winsamp_shift
            while (j + winsamp < sig_len):
                win = sig[j:j + winsamp]
                count = count + 1
                windowed_signal.append(win)
                j += winsamp_shift
            windowed_signal = np.vstack(windowed_signal)
        else:  # in the case that overlap is false (win_shift = 0)
            j += winsamp
            while (j + winsamp < sig_len):
                win = sig[j:j + winsamp]
                count = count + 1
                windowed_signal.append(win)
                j += winsamp
            windowed_signal = np.vstack(windowed_signal)
        signals.append(windowed_signal)
        # for i in range(count):
        #     # append the features to the list of features,
        #     # as these features are the same for all segmented windowed signals
        #     features.append(current_features)
    # features = np.vstack(features)
    signals = np.vstack(signals)
    # print('type of features:',type(features))
    # features = feature_scaler.transform(features)
    signals = signal_scaler.transform(signals)
    dim = signals.shape
    signals = signals.reshape(dim[0], dim[1], 1)  # for feeding the CNN model
    # print('shape of signals:',signals.shape)
    # print('shape of features:',features.shape)

    # Get classifier probabilities: wide & deep
    # probabilities = classifier.predict({'signal_input':signals, 'feature_input': features})

    # Get classifier probabilities: deep (not wide)
    murmur_probabilities = murmur_classifier.predict(signals)
    murmur_probabilities = np.max(np.asarray(murmur_probabilities, dtype=np.float32), axis=0)
    murmur_probabilities = murmur_probabilities / np.sum(murmur_probabilities)

    outcome_probabilities = outcome_classifier.predict(signals)
    outcome_probabilities = np.max(np.asarray(outcome_probabilities, dtype=np.float32), axis=0)
    outcome_probabilities = outcome_probabilities / np.sum(outcome_probabilities)

    # Choose label with higher probability.
    murmur_labels = np.zeros(len(murmur_classes), dtype=np.int_)
    idx = np.argmax(murmur_probabilities)
    murmur_labels[idx] = 1

    outcome_labels = np.zeros(len(outcome_classes), dtype=np.int_)
    idx = np.argmax(outcome_probabilities)
    outcome_labels[idx] = 1

    # Concatenate classes, labels, and probabilities.
    classes = murmur_classes + outcome_classes
    labels = np.concatenate((murmur_labels, outcome_labels))
    probabilities = np.concatenate((murmur_probabilities, outcome_probabilities))

    return classes, labels, probabilities

################################################################################
#
# Optional functions. You can change or remove these functions and/or add new functions.
#
################################################################################

# Save your trained model.
# def save_challenge_model(model_folder, classes, classifier, signal_scaler, feature_scaler):
def save_challenge_model(model_folder, murmur_classes, outcome_classes, murmur_classifier, outcome_classifier,
                         signal_scaler):
    # filename = os.path.join(model_folder, 'model.sav')
    filename = model_folder
    # Construct a data structure for the model and save it.
    murmur_classifier.save(filename + '/murmur_model')
    outcome_classifier.save(filename + '/outcome_model')
    pd.DataFrame(murmur_classes).to_csv(filename + '/murmur_classes.csv', sep=',', index=False)
    pd.DataFrame(outcome_classes).to_csv(filename + '/outcome_classes.csv', sep=',', index=False)
    # scaler = {'signal_scaler': signal_scaler, 'feature_scaler': feature_scaler}
    scaler = {'signal_scaler': signal_scaler}
    pickle.dump(scaler, open(filename + '/scaler.pkl', 'wb'))


# Extract features from the data.
def get_features(data, recordings):
    # Extract the age group and replace with the (approximate) number of months for the middle of the age group.
    age_group = get_age(data)

    if compare_strings(age_group, 'Neonate'):
        age = 0.5
    elif compare_strings(age_group, 'Infant'):
        age = 6
    elif compare_strings(age_group, 'Child'):
        age = 6 * 12
    elif compare_strings(age_group, 'Adolescent'):
        age = 15 * 12
    elif compare_strings(age_group, 'Young Adult'):
        age = 20 * 12
    else:
        age = float('nan')

    # Extract sex. Use one-hot encoding.
    sex = get_sex(data)

    sex_features = np.zeros(2, dtype=int)
    if compare_strings(sex, 'Female'):
        sex_features[0] = 1
    elif compare_strings(sex, 'Male'):
        sex_features[1] = 1

    # Extract height and weight.
    height = get_height(data)
    weight = get_weight(data)

    # Extract pregnancy status.
    is_pregnant = get_pregnancy_status(data)

    # Extract recording locations and data. Identify when a location is present, and compute
    # the mean, variance, and skewness of each recording. If there are multiple recordings for one location,
    # then extract features from the last recording.
    locations = get_locations(data)

    recording_locations = ['AV', 'MV', 'PV', 'TV', 'PhC']
    num_recording_locations = len(recording_locations)
    recording_features = np.zeros((num_recording_locations, 4), dtype=float)
    num_locations = len(locations)
    num_recordings = len(recordings)
    if num_locations == num_recordings:
        for i in range(num_locations):
            for j in range(num_recording_locations):
                if compare_strings(locations[i], recording_locations[j]) and np.size(recordings[i]) > 0:
                    recording_features[j, 0] = 1
                    recording_features[j, 1] = np.mean(recordings[i])
                    recording_features[j, 2] = np.var(recordings[i])
                    recording_features[j, 3] = sp.stats.skew(recordings[i])
    recording_features = recording_features.flatten()

    features = np.hstack(([age], sex_features, [height], [weight], [is_pregnant], recording_features))

    return np.asarray(features, dtype=np.float32)
